"""
(c) 2020 АО 'НИИ "МАСШТАБ"'
"""

import logging

from .base_object import VeilApiObject, asyncable
from .node import Node

LOG = logging.getLogger("bazalt")


class Domain(VeilApiObject):

    base_url = "domains/"

    @classmethod
    @asyncable
    def create(cls, verbose_name, node, cpu_count=1, memory_count=256,
               description="", context=None, os_type='Linux'):
        r = cls.api_session.post(cls.prepare_url(cls.base_url, context), data={
            "node": node,
            "cpu_count": cpu_count,
            "memory_count": memory_count,
            "description": description,
            "verbose_name": verbose_name,
            "os_type": os_type})
        r.raise_for_status_code()
        LOG.info("Domain {} create".format(verbose_name))
        return r, cls

    @classmethod
    @asyncable
    def multi_create(cls, domain_configuration, context=None):
        r = cls.api_session.post(cls.prepare_url(cls.base_url + "multi-create-domain/", context),
                                 data=domain_configuration)
        r.raise_for_status_code()
        LOG.info("Domain {} multi_create".format(domain_configuration.get('verbose_name', '?')))
        return r, cls

    @asyncable
    def delete(self, force=False, full=False, context=None):
        LOG.info("Domain {} delete".format(self.name))
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + "remove/", context),
                                  data={"force": force, "full": full})
        r.raise_for_status_code()
        return r, None

    @classmethod
    def convert_from_xml(cls, xml_sources, context=None):
        r = cls.api_session.post(cls.prepare_url(cls.base_url + "convert-from-xml/", context),
                                 data=xml_sources)
        r.raise_for_status_code()
        LOG.info("Xml converted to domain(s)")
        return r

    @classmethod
    def show_backup(cls, file, context=None):
        r = cls.api_session.post(cls.prepare_url(cls.base_url + "show-backup/", context),
                                 data={'file': file})
        r.raise_for_status_code()
        LOG.info("Show backup {}.".format(file))
        return r.data

    @classmethod
    def config_restore(cls, file, node=None, datapool=None, context=None):
        data = {'file': file}
        if node:
            data.update({'node': node})
        if datapool:
            data.update({'datapool': datapool})
        r = cls.api_session.post(cls.prepare_url(cls.base_url + "config-restore/", context),
                                 data=data)
        r.raise_for_status_code()
        LOG.info("Got config for restore from backup {}.".format(file))
        return r.data

    @classmethod
    @asyncable
    def automated_restore(cls, file, node, datapool=None, context=None):
        data = {'file': file, 'node': node}
        if datapool:
            data.update({'datapool': datapool})
        r = cls.api_session.post(cls.prepare_url(cls.base_url + "automated-restore/", context),
                                 data=data)
        r.raise_for_status_code()
        LOG.info("Automated restore from {}.".format(file))
        return r, cls

    @asyncable
    def start(self, context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + "start/", context))
        r.raise_for_status_code()
        LOG.info("Domain {} start".format(self.name))
        return r, None

    @asyncable
    def shutdown(self, force=True, context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + "shutdown/", context),
                                  data={"force": force})
        r.raise_for_status_code()
        LOG.info("Domain {} shutdown".format(self.name))
        return r, None

    @asyncable
    def suspend(self, context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + "suspend/", context))
        r.raise_for_status_code()
        LOG.info("Domain {} suspend".format(self.name))
        return r, None

    @asyncable
    def resume(self, context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + "resume/", context))
        r.raise_for_status_code()
        LOG.info("Domain {} resume".format(self.name))
        return r, None

    @asyncable
    def reboot(self, force=True, context=None):
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + "reboot/", context),
                                  data={"force": force})
        r.raise_for_status_code()
        LOG.info("Domain {} reboot".format(self.name))
        return r, None

    @classmethod
    def multi_manager(cls, action, entity_ids, options=None, context=None):
        r = cls.api_session.post(cls.prepare_url(cls.base_url + "multi-manager/", context),
                                 data={"action": action,
                                       "entity_ids": entity_ids,
                                       "options": options if options else {}})
        if context and context.get("dry-run"):
            return r
        r.raise_for_status_code()
        LOG.info("Multi manager action {} for domains: {}".format(action, ', '.join(entity_ids)))
        return r, None

    def attach_vdisk(self, vdisk, target_bus=None, driver_cache=None, **kwargs):
        data = {"vdisk": vdisk}
        if target_bus is not None:
            data['target_bus'] = target_bus
        if driver_cache is not None:
            data['driver_cache'] = driver_cache
        self.simple_instance_action("attach-vdisk", data=data, **kwargs)

    def detach_vdisk(self, vdisk, **kwargs):
        return self.simple_instance_action("detach-vdisk", data={"vdisk": vdisk}, **kwargs)

    def attach_lun(self, lun, target_bus=None, driver_cache=None, **kwargs):
        data = {"lun": lun}
        if target_bus is not None:
            data['target_bus'] = target_bus
        if driver_cache is not None:
            data['driver_cache'] = driver_cache
        self.simple_instance_action("attach-lun", data=data, **kwargs)

    def detach_lun(self, lun, **kwargs):
        return self.simple_instance_action("detach-lun", data={"lun": lun}, **kwargs)

    def attach_iso(self, iso, cdrom):
        return self.simple_instance_action("attach-iso", data={"iso": iso, "cdrom": cdrom})

    def detach_iso(self, cdrom):
        return self.simple_instance_action("detach-iso", data={"cdrom": cdrom})

    def attach_usb(self, usb, usb_controller=0, action_type="usb_devices"):
        if action_type == "usb_devices":
            return self.simple_instance_action("attach-usb",
                                               data={"usb": usb,
                                                     "usb_controller": usb_controller,
                                                     "action_type": action_type})
        elif action_type == "spice_usb_channel":
            return self.simple_instance_action("attach-usb",
                                               data={"usb_controller": usb_controller,
                                                     "action_type": action_type})

    def detach_usb(self, usb, controller_order=0, action_type="usb_device"):
        if action_type == "usb_device":
            return self.simple_instance_action("detach-usb", data={"usb": usb,
                                                                   "action_type": action_type})
        elif action_type == "spice_usb_channel":
            return self.simple_instance_action("detach-usb", data={"controller_order": controller_order,
                                                                   "action_type": action_type})

    def attach_pci(self, **kwargs):
        """

        :param kwargs: {'pcis': list of dicts, 'pci_controller(optional)': {'model': string, 'order': int}}
        """
        return self.simple_instance_action("attach-pci",
                                           data=kwargs)

    def detach_pci(self, pci):
        return self.simple_instance_action("detach-pci", data={"pci": pci})

    def attach_mdev(self, **kwargs):
        """
        :param kwargs: {'pci_id' : core_pci_id,
                        'md_type': mdev_type,
                        'display': 'on'/'off'}
        API='vfio-pci' is hardcoded as for now
        :return:
        """
        return self.simple_instance_action('attach-mdev', data=kwargs)

    def detach_mdev(self, _uuid):
        return self.simple_instance_action('detach-mdev', data={'uuid': _uuid})

    @asyncable
    def clone(self, node=None, verbose_name=None, datapool=None, snapshot=None, context=None):
        data = {}
        if node:
            data['node'] = node
        if verbose_name:
            data['verbose_name'] = verbose_name
        if datapool:
            data['datapool'] = datapool
        if snapshot:
            data['snapshot'] = snapshot
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + 'clone/', context), data=data)
        r.raise_for_status_code()
        return r, self.__class__

    @classmethod
    def free_domain_name(cls, verbose_name=None, parent=None, context=None):
        data = {}
        if verbose_name:
            data['verbose_name'] = verbose_name
        if parent:
            data['parent'] = parent
        r = cls.api_session.post(cls.prepare_url(cls.base_url + "free-domain-name/", context),
                                 data=data)
        r.raise_for_status_code()
        return r.data

    def update_controllers(self, **controllers):
        return self.simple_instance_action("controllers", data=controllers)

    def update(self, **res):
        r = self.api_session.put(self.get_self_url(self.uuid), data=res)
        r.raise_for_status_code()

    def update_ha_options(self, **res):
        r = self.api_session.put(self.get_self_url(self.uuid) + 'update-ha-options/', data=res)
        r.raise_for_status_code()

    def update_disastery_options(self, **res):
        r = self.api_session.put(self.get_self_url(self.uuid) + 'update-disastery-options/', data=res)
        r.raise_for_status_code()

    def update_safety(self, **res):
        r = self.api_session.put(self.get_self_url(self.uuid) + 'safety/', data=res)
        r.raise_for_status_code()

    def template(self, template):
        data = {'template': template}
        r = self.api_session.put(self.get_self_url(self.uuid) + 'template/', data=data)
        r.raise_for_status_code()
        return r

    def update_resources(self, **resources):
        return self.simple_instance_action("resources", data=resources)

    def update_video(self, **video):
        return self.simple_instance_action("video", data=video)

    def update_sound(self, **sound):
        return self.simple_instance_action("sound", data=sound)

    @asyncable
    def migrate(self, target_node=None, non_shared=None, target_datapool=None, context=None):
        data = {}
        if target_node is not None:
            data['target_node'] = target_node
        if non_shared is not None:
            data['non_shared'] = non_shared
        if target_datapool is not None:
            data['target_datapool'] = target_datapool
        r = self.api_session.post(self.prepare_url(self.get_self_url(self.uuid) + "migrate/", context),
                                  data=data)
        r.raise_for_status_code()
        LOG.info("Domain {} migrate".format(self.name))
        return r, None

    def migrate_ready_nodes(self, non_shared=False):
        base_url = self.get_self_url(self.uuid) + 'migrate-available-nodes/'
        if non_shared:
            base_url += '?non_shared=true'
        r = self.api_session.get(base_url, params={})
        r.raise_for_status_code()
        return r.data['nodes']

    def recover(self, target_node):
        data = dict(target_node=target_node)
        return self.simple_instance_action("recover", data=data)

    def check_recover(self, target_node):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'check-recover/',
                                 params={'target_node': target_node})
        r.raise_for_status_code()
        return r.data

    def recover_ready_nodes(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'recover-ready-nodes/',
                                 params={})
        r.raise_for_status_code()
        return r.data['nodes']

    def spice(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'spice/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    def os_types(self):
        r = self.api_session.get(self.base_url + 'os-types/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    def os_defaults(self):
        r = self.api_session.get(self.base_url + 'os-defaults/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    def statistics(self, request_time=1):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'statistics/?time={}'.format(request_time),
                                 params={})
        r.raise_for_status_code()
        return r.data

    def stats_v2(self, start, end, step=1):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'stats-v2/?start={}&end={}&step={}'.format(start,
                                                                                                           end,
                                                                                                           step),
                                 params={})
        r.raise_for_status_code()
        return r.data

    def set_owners(self, owners):
        r = self.api_session.put(self.get_self_url(self.uuid) + 'set-owners/',
                                 data={'owners': owners})
        r.raise_for_status_code()

    def update_graphics_password(self, graphics_password):
        r = self.api_session.put(self.get_self_url(self.uuid) + 'update-graphics-password/',
                                 data={'graphics_password': graphics_password})
        r.raise_for_status_code()

    def update_graphics_settings(self, graphics_settings):
        r = self.api_session.put(self.get_self_url(self.uuid) + 'update-graphics-settings/',
                                 data=graphics_settings)
        r.raise_for_status_code()

    def update_remote_access(self, remote_access):
        return self.simple_instance_action("remote-access", data={'remote_access':
                                                                  remote_access})

    def update_remote_access_white_list(self, remote_access_allow_all,
                                        ip_addr, comment='default'):
        return self.simple_instance_action("remote-access-white-list",
                                           data={'remote_access_allow_all': remote_access_allow_all,
                                                 'remote_access_white_list': [
                                                     {'ip_addr': ip_addr,
                                                      'comment': comment}]})

    def update_boot_order(self, devices):
        r = self.api_session.put(self.get_self_url(self.uuid) + 'update-boot/',
                                 data={'devices': devices})
        r.raise_for_status_code()

    def boot(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'boot/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    def boot_free(self):
        r = self.api_session.get(self.get_self_url(self.uuid) + 'boot-free/',
                                 params={})
        r.raise_for_status_code()
        return r.data

    def cloud_init(self, cloud_init_data):
        r = self.api_session.put(self.get_self_url(self.uuid) + 'cloud-init/',
                                 data=cloud_init_data)
        r.raise_for_status_code()

    @property
    def node(self):
        return Node.get(self.data.node['id'])

    @property
    def name(self):
        return self.data.verbose_name

    @property
    def is_powered(self):
        return self.data.user_power_state >= 2

    @property
    def suspended(self):
        return self.data.user_power_state == 2

    @property
    def cpu_count(self):
        if 'cpu_topology' not in self.data:
            self.refresh()
        return self.data.cpu_topology['cpu_count']

    @property
    def cpu_map(self):
        if 'cpu_topology' not in self.data:
            self.refresh()
        return self.data.cpu_topology['cpu_map']
