"""
(c) 2020 АО 'НИИ "МАСШТАБ"'
"""
import warnings
with warnings.catch_warnings():
    warnings.simplefilter('ignore', category=DeprecationWarning)
    from collections import namedtuple
    from typing import Iterable
import json
import logging
import requests
import urllib3

from veilrpc.utils.bazalt import retry_on_exception

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

LOG = logging.getLogger("bazalt")

RequestType = namedtuple('RequestType', 'GET POST PUT')('get', 'post', 'put')
VEIL_API_REQ_PROTO = "http://"
VEIL_API_WS_PROTO = "ws://"
JWT_TOKEN_PREFIX = "JWT"  # nosec
VEIL_API_HOST_SSL_VERIFY = False
VEIL_LANGUAGE = "ru"


class ApiError(Exception):
    pass


class ApiAuthError(ApiError):
    pass


class ApiConnectionError(ApiError):
    pass


class ApiResponse:
    def __init__(self, request_type, response):
        self._response = response
        self.request_type = request_type
        LOG.debug(f"API response. status: {self.status_code} data: {self._response}")

    @property
    def uuid(self):
        return self._get_response_filed("id")

    @property
    def status_code(self):
        return self._response.status_code

    @property
    def async_response(self):
        return True if self.status_code == 202 else False

    @property
    def data(self):
        try:
            return self._response.json()
        except json.JSONDecodeError as e:
            raise ApiError(f"invalid API response: {e}") from None

    @property
    def task_uuid(self):
        task_field = self._get_response_filed("_task")
        task_id = self._get_field(task_field, "id")
        return task_id

    @property
    def list_results(self):
        return self._get_response_filed("results")

    @property
    def count_results(self):
        return self._get_response_filed("count")

    def raise_for_status_code(self, expected_status=(200, 202)):
        LOG.debug(f"response status: {self.status_code}")
        if not isinstance(expected_status, Iterable):
            expected_status = [expected_status]
        if self.status_code not in expected_status:
            try:
                data = self._response.json()
            except json.JSONDecodeError:
                data = self._response.content.decode()
            raise ApiError(
                f"unexpected status code. "
                f"expected: {expected_status} actual: {self.status_code} data: {data} url: {self._response.url}")

    def _get_response_filed(self, key):
        return self._get_field(self.data, key)

    @staticmethod
    def _get_field(data, key):
        try:
            return data[key]
        except KeyError as err:
            raise ApiError(f"no {key} field in response") from err


class ApiSession:

    def __init__(self,
                 controller_ip,
                 username=None,
                 password=None,
                 token=None,
                 ldap=False,
                 ssl=VEIL_API_HOST_SSL_VERIFY,
                 proto=VEIL_API_REQ_PROTO,
                 auth_url='auth',
                 url_prefix='api',
                 logout_url='logout'):
        if not ((username and password) or token):
            raise ApiConnectionError('username and password or token are required')
        self.username = username
        self.password = password
        self.token = token
        self.ldap = ldap
        self._session = None
        self.verify_ssl = ssl
        self.proto_host = f'{proto}{controller_ip}'
        self.auth_url = f'{self.proto_host}/{auth_url}/'
        self.url_prefix = f'{self.proto_host}/{url_prefix}/'
        self.logout_url = f'{self.proto_host}/{logout_url}/'

    def _reconnect(self, k):
        if k.status_code == 401:
            self.refresh_session_token()
            # pylint: disable = protected-access
            LOG.warning("Session token invalid. Refreshing. url: %s", k._response.url)
            raise requests.exceptions.ConnectionError
        return k

    @retry_on_exception(requests.exceptions.ConnectionError)
    def get(self, url, params=None, **kwargs):
        url = self._add_url_prefix_if_needed(url)
        LOG.debug(f"get request. url: {url}, params: {params}")
        session = self._get_session()
        return self._reconnect(ApiResponse(RequestType.GET, session.get(url, params=params,
                                                                        **kwargs)))

    @retry_on_exception(requests.exceptions.ConnectionError)
    def post(self, url, asynced=False, data=None, **kwargs):
        url = self._add_url_prefix_if_needed(url)
        if asynced:
            if url.endswith('/'):
                url += "?async=1"
            else:
                url += "/?async=1"
        LOG.debug("post request. url: %s, data: %s", url, data)
        session = self._get_session()
        return self._reconnect(ApiResponse(RequestType.POST,
                               session.post(url, json=data, **kwargs)))

    @retry_on_exception(requests.exceptions.ConnectionError)
    def put(self, url, data=None, **kwargs):
        url = self._add_url_prefix_if_needed(url)
        LOG.debug("put request. url: %s, data: %s", url, data)
        session = self._get_session()
        return self._reconnect(ApiResponse(RequestType.PUT,
                               session.put(url, json=data, **kwargs)))

    def get_raw_requests_session(self):
        return self._get_session()

    def refresh_session_token(self):
        return self._create_new_session()

    def _get_session(self):
        if not self._session:
            self._create_new_session()
        return self._session

    def _create_new_session(self):
        s = requests.Session()
        s.headers.update(
            {"Authorization": f"{JWT_TOKEN_PREFIX} {self._get_token()}",
             "Content-Type": "application/json; charset=utf8",
             "Accept-Language": VEIL_LANGUAGE})
        s.verify = self.verify_ssl
        self._session = s
        LOG.info("successfully AUTH at VEIL API")

    @retry_on_exception(requests.exceptions.ConnectionError)
    def _get_token(self):
        if self.token:
            return self.token

        def connect_legacy():
            data = {'username': self.username, 'password': self.password, 'ldap': self.ldap}
            try:
                # pylint: disable = missing-timeout
                return requests.post(self.auth_url, data=data, verify=self.verify_ssl)
            except requests.exceptions.ConnectionError as e:
                raise ApiConnectionError("can't connect to VEIL API: ", e) from None

        response = connect_legacy()

        try:
            response = json.loads(response.text)
            return response['token']
        except (json.JSONDecodeError, KeyError, TypeError) as err:
            raise ApiAuthError(f"unexpected response from VEIL API: {response}") from err

    def _add_url_prefix_if_needed(self, url):
        return f"{self.url_prefix}{url}" if self.url_prefix else url
