"""
(c) 2020 АО 'НИИ "МАСШТАБ"'
"""

import logging
import time
from typing import Iterable

from veilrpc.utils.api.session import ApiError
from veilrpc.utils.api.base_object import VeilApiObject, asyncable


LOG = logging.getLogger("bazalt")


class TaskStatus:
    IN_PROGRESS = ["IN_PROGRESS", "В процессе", "В ПРОЦЕССЕ"]
    SUCCESS = ["SUCCESS", "Выполнена", "ВЫПОЛНЕНА"]
    FAILED = ["FAILED", "Не выполнена", "НЕ ВЫПОЛНЕНА"]
    CANCELED = ["CANCELED", "Пользовательский выход", "ПОЛЬЗОВАТЕЛЬСКИЙ ВЫХОД"]
    LOST = ["LOST", "Утеряна", "УТЕРЯНА"]
    PARTIAL = ["PARTIAL", "Частичное выполнение", "ЧАСТИЧНОЕ ВЫПОЛНЕНИЕ"]


def wait_task_ready(uuid, blocked=True, timeout=None, raise_timeout=True):
    timeout = timeout or 300
    task = Task.get(uuid)
    task.refresh()
    if task.in_final_status:
        return task
    elif blocked:
        timeout_time = int(time.monotonic()) + timeout
        while True:
            task.refresh()
            if task.in_final_status:
                time.sleep(1)
                task.refresh()
                return task
            if int(time.monotonic()) > timeout_time:
                if raise_timeout:
                    if task.data.is_multitask:
                        childs = str([t.name for t in task.childs()])
                    else:
                        childs = ''
                    msg = "Task: {} wait {}s timeout error. is_multitask: {}. progress: {}. status: {}. childs: {}"
                    raise ApiError(msg.format(task.name, timeout, str(task.data.is_multitask),
                                   str(task.data.progress), task.data.status, childs))
                else:
                    return task
            time.sleep(1)


class Task(VeilApiObject):

    base_url = "tasks/"

    @property
    def status(self):
        return self.data.status.upper()

    @property
    def in_final_status(self):
        return self.status not in TaskStatus.IN_PROGRESS

    @property
    def successful(self):
        return self.status in TaskStatus.SUCCESS

    @property
    def failure(self):
        return self.status in TaskStatus.FAILED

    @property
    def in_progress(self):
        return self.status in TaskStatus.IN_PROGRESS

    @property
    def partial(self):
        return self.status in TaskStatus.PARTIAL

    @property
    def not_executed(self):
        return self.status in TaskStatus.CANCELED or self.status in TaskStatus.LOST

    @property
    def lost(self):
        return self.status in TaskStatus.LOST

    @property
    def canceled(self):
        return self.status in TaskStatus.CANCELED

    def raise_for_status(self, status=TaskStatus.SUCCESS[1]):
        if not isinstance(status, Iterable):
            status = [status]
        if self.status not in status:
            if self.data.is_multitask:
                childs = str([t for t in self.childs()])
            else:
                childs = ''
            raise ApiError(
                "unexpected task status."
                "expected: {} actual: {} data: {} progress: {} childs: {}".format(status,
                                                                                  self.status,
                                                                                  self.data,
                                                                                  str(self.data.progress),
                                                                                  childs
                                                                                  ))

    def raise_for_status_success(self):
        if self.status not in TaskStatus.SUCCESS:
            raise ApiError(
                "unexpected task status."
                "expected: {} actual: {} data: {}".format(', '.join(TaskStatus.SUCCESS),
                                                          self.status,
                                                          self.data))

    @property
    def response(self):
        r = self.api_session.get("{}{}/".format(self.get_self_url(self.uuid), "response"))
        r.raise_for_status_code()
        return r.data

    @property
    def jid(self):
        r = self.api_session.get("{}{}/".format(self.get_self_url(self.uuid), "jid"))
        r.raise_for_status_code()
        return r.data['jid']

    def release_locks(self):
        r = self.api_session.put(self.get_self_url(self.uuid) + 'release-locks/',
                                 data={})
        r.raise_for_status_code()

    @classmethod
    def check(cls):
        LOG.info("Tasks check")
        r = cls.api_session.put("{url}{action}/".format(
            url=cls.base_url,
            action="check"))
        r.raise_for_status_code()

    def wait(self, **kwargs):
        wait_task_ready(uuid=self.uuid, **kwargs)
        self.refresh()

    @asyncable
    def cancel(self, force=False, context=None):
        r = self.api_session.put(self.prepare_url(self.get_self_url(self.uuid) + "cancel/",
                                                  context),
                                 data={"force": force})
        return r, None

    def childs(self):
        r = self.api_session.get(self.base_url, params={"parent": self.uuid})
        r.raise_for_status_code()
        result = []
        for t in r.list_results:
            result.append(self.__class__(t))
        return result
